import numpy as np
from rangel import RangeCollection

def rasterize_events(
    events,
    values=None,
    size=1,
    blur=0,
    blur_style='linear',
    normalize=True,
    bounds=None,
    fill='cut', 
    closed='left_mod',
    rc=None,
    **kwargs
):
    """
    Digitize and buffer events over a defined range, extending them across 
    uniform steps within the range and scaling their values relative to their 
    digital distance from their intersecting step location.
    
    Parameters
    ----------
    events : 1d or 2d array-like
        The locations of the target events being analyzed. If points, must be 
        flat array-like or of shape (x, 1). If ranges, must be array-like of 
        shape (x, 2). All event locations should fall on the range [beg, end] 
        (if provided) to be considered in the analysis.
    values : numeric or 1d array-like, optional
        The value(s) associated with each event being analyzed. If not 
        provided, all values will default to be 1.
    size : positive numeric, default 1
        The length of each pixel used to perform the analysis.
    blur : int, default 0
        The number of pixels to blur events across based on the blur style.
    blur_style : str or callable, default 'linear'
        The scaling function to be called at each blurring step to scale 
        original values. If a callable is provided, it must accept a single 
        integer input for the zero-indexed pixel number, returning a single 
        float scaling value. Predefined blurring functions can be called using 
        the following labels:
        
        Options
        -------
        linear : linearly scale down values from the original value to zero 
            at the first index outside the blurred pixel range
        none : do not scale down original values
        
    normalize : bool, default True
        Whether to normalize resulting buffered values so each event's total 
        buffers sum to 1.
    bounds : two-value numeric tuple, optional
        The values at which to begin and end the pixelation analysis. If not 
        provided, will default to the min and max location values in the events 
        data, respectively. If a predefined range collection is provided, that 
        will supersede these input parameters.
    fill : {'none','cut','left','right'}, default 'cut'
        How to fill a gap at the end of an event's range.

        Options
        -------
        none : no window will be generated to fill the gap at the end of 
            the input range.
        cut : a truncated window will be created to fill the gap with a 
            length less than the full window length.
        left : the final window will be anchored on the end of the event  
            and will extend the full window length to the left. 
        right : the final window will be anchored on the grid defined by 
            the step value, extending the full window length to the right, 
            beyond the event's end value.
    
    closed : str {'left', 'left_mod', 'right', 'right_mod', 'both', 
            'neither'}, default 'left'
        Whether window intervals are closed on the left-side, right-side, 
        both, or neither. Values of 'left' and 'right' will avoid double-
        counting and under-counting events which occur on the edges of 
        windows. If a predefined range collection is provided, that will 
        supersede these input parameters.

        Options
        -------
        left : ranges are always closed on the left and never closed on the 
            right.
        left_mod : ranges are always closed on the left and only closed on the 
            right when the next range is not consecutive.
        right : ranges are always closed on the right and never closed on the 
            right.
        right_mod : ranges are always closed on the right and only closed on 
            the left when the previous range is not consecutive.
        both : ranges are always closed on both sides
        neither : ranges are never closed on either side
        
    rc : RangeCollection, optional
        A predefined valid consecutive range collection instance which defines 
        the analysis domain and which will be used to intersect with input 
        events.
            
    Created:  2022-01-06
    Modified: 2022-01-12
    """
    ##################
    # VALIDATE INPUT #
    ##################
    
    # Validate blur style
    if blur_style=='linear':
        blur_function = lambda n: (blur - n) / blur
    elif blur_style=='none':
        blur_function = lambda n: 1
    elif callable(blur_style):
        blur_function = blur_style
    else:
        raise ValueError("Input blur_style must be callable or label of "
                            "valid predefined scaling function.")            
    
    # Validate events
    try:
        events = np.asarray(events, dtype=float)
    except ValueError:
        raise TypeError("Could not convert input events to np.ndarray of "
                        "dtype float.")
    # - Determine shape and event type
    if events.ndim == 1:
        events = np.append(events, events).reshape(2,-1).T
    elif events.ndim == 2 and events.shape[1] == 1:
        events = np.append(events, events).reshape(2,-1).T
    elif events.ndim == 2 and events.shape[1] == 2:
        pass
    else:
        raise ValueError("Invalid events input. Must be array-like of "
                         "shape (x,), (x, 1), or (x, 2).")
    
    # Validate values data
    if not values is None:
        try:
            values = np.full(events.shape[0], float(values)) # extend scalar
        except:
            try:
                values = np.asarray(values, dtype=float).flatten() # coerce
            except:
                raise TypeError("Could not convert input values to np.ndarray "
                                "of dtype float.")
        if not values.shape[0] == events.shape[0]:
            raise ValueError("Number of event values must be equal to the "
                            "number of events provided. Provided: "
                            f"{events.shape[0]} events, {values.shape[0]} "
                            "values.")
    
    # Validate analysis range collection
    if rc is None:
        if bounds is None:
            beg = np.min(events) if bounds is None else bounds[0]
            end = np.max(events) if bounds is None else bounds
        else:
            try:
                beg, end = bounds
            except:
                raise ValueError("If used, bounds must be provided as "
                                 "two-value tuple of scalars defining the "
                                 "bounds of the analysis.")
        # Create a new analysis range collection
        rc = RangeCollection.from_steps(beg, end, length=size, steps=1, 
                                        fill=fill, closed=closed)
    elif not isinstance(rc, RangeCollection):
        raise ValueError("Input range collection is not valid.")
            
    ####################
    # PERFORM ANALYSIS #
    ####################
    
    # Intersect events with analysis range collection
    intx = rc.is_intersecting(beg=events[:,0],
                              end=events[:,1],
                              squeeze=False) * 1
    
    # Blur events
    scale = blur_function(0)
    data = intx * scale
    for step in range(1, blur):
        # Create and blur buffered data
        scale = blur_function(step)
        buff = np.pad(intx, ((step, step), (0, 0)), mode='constant') * scale
        # Apply buffered data
        forw = buff[:-step * 2, :]
        back = buff[step * 2:, :]
        data += forw + back
        
    # Normalize buffered data
    if normalize:
        denom = data.sum(axis=0)
        data = np.divide(data, denom, out=np.zeros_like(data), where=denom!=0)
    
    # Apply values if used
    if not values is None:
        return np.multiply(data, values)
    else:
        return data

def buffer_events(
    events, 
    size, 
    steps, 
    scaler='linear', 
    normalize=True, 
    beg=None, 
    end=None, 
    fill='cut', 
    closed='left_mod', 
    rc=None,
    values=None, # ADDED
):
    """
    Digitize and buffer events over a defined range, extending them across 
    uniform steps within the range and scaling their values relative to their 
    digital distance from their intersecting step location.
    
    Parameters
    ----------
    events : 1d or 2d array-like
        The locations of the target events being analyzed. If points, must be 
        flat array-like or of shape (x, 1). If ranges, must be array-like of 
        shape (x, 2). All event locations should fall on the range [beg, end] 
        (if provided) to be considered in the analysis.
    size : positive numeric
        The size of each buffering step used to perform the analysis.
    steps : int, default 10
        The number of steps of buffering to perform.
    scaler : str or callable, default 'linear'
        The scaling function to be called at each buffering step to scale 
        buffered values. If a callable is provided, it must accept a single 
        integer input for the zero-indexed buffering step, returning a single 
        float scaling value. Predefined scaling functions can be called using 
        the following labels:
        
        Options
        -------
        linear : linearly scale down values from the center value to the 
            edge value
        none : do not scale down buffered values
        
    normalize : bool, default True
        Whether to normalize resulting buffered values so each event's total 
        buffers sum to 1.
    beg, end : numeric, optional
        The locations at which to begin and end the buffering analysis. 
        If not provided, will default to the min and max values in the 
        events data, respectively. If a predefined range collection is 
        provided, that will supersede these input parameters.
    closed : str {'left', 'left_mod', 'right', 'right_mod', 'both', 
            'neither'}, default 'left'
        Whether window intervals are closed on the left-side, right-side, 
        both, or neither. Values of 'left' and 'right' will avoid double-
        counting and under-counting events which occur on the edges of 
        windows.

        Options
        -------
        left : ranges are always closed on the left and never closed on the 
            right.
        left_mod : ranges are always closed on the left and only closed on the 
            right when the next range is not consecutive.
        right : ranges are always closed on the right and never closed on the 
            right.
        right_mod : ranges are always closed on the right and only closed on 
            the left when the previous range is not consecutive.
        both : ranges are always closed on both sides
        neither : ranges are never closed on either side
        
    rc : RangeCollection, optional
        A predefined valid consecutive range collection instance which defines 
        the analysis domain and which will be used to intersect with input 
        events.
            
    Created:  2022-01-06
    Modified: 2022-01-06
    """
    ##################
    # VALIDATE INPUT #
    ##################
    
    # Validate scaler
    if scaler=='linear':
        scaler = lambda n: (steps - n) / steps
    elif scaler=='none':
        scaler = lambda n: 1
    else:
        if not callable(scaler):
            raise ValueError("Input scaler must be callable or label of valid "
                             "predefined scaler function.")
    
    # Validate events
    try:
        events = np.asarray(events, dtype=float)
    except ValueError:
        raise TypeError("Could not convert input events to np.ndarray of "
                        "dtype float.")
    # - Determine shape and event type
    if events.ndim == 1:
        events = np.append(events, events).reshape(2,-1).T
    elif events.ndim == 2 and events.shape[1] == 1:
        events = np.append(events, events).reshape(2,-1).T
    elif events.ndim == 2 and events.shape[1] == 2:
        pass
    else:
        raise ValueError("Invalid events input. Must be array-like of "
                         "shape (x,), (x, 1), or (x, 2).")
    
    # Validate values data
    if values is None:
        values = 1 # default
    try:
        values = np.full(events.shape[0], float(values)) # extend single value
    except:
        try:
            values = np.asarray(values, dtype=float).flatten() # coerce
        except:
            raise TypeError("Could not convert input values to np.ndarray "
                            "of dtype float.")
    if not values.shape[0] == events.shape[0]:
        raise ValueError("Number of event values must be equal to the number "
                         f"of events provided. Provided: {events.shape[0]} "
                         f"events, {values.shape[0]} values.")
            
    ####################
    # PERFORM ANALYSIS #
    ####################
    
    # Intersect events with range
    if rc is None:
        beg = np.min(events) if beg is None else beg
        end = np.max(events) if end is None else end
        rc = RangeCollection.from_steps(beg, end, length=size, steps=1, 
                                        fill=fill, closed=closed)
    elif not isinstance(rc, RangeCollection):
        raise ValueError("Input range collection is not valid.")
    intx = rc.is_intersecting(beg=events[:,0], end=events[:,1]) * 1
    
    # Buffer events
    scale = scaler(0)
    data = intx * scale
    for step in range(1, steps):
        # Create and scale buffered data
        scale = scaler(step)
        buff = np.pad(intx, ((step, step), (0, 0)), mode='constant') * scale
        # Apply buffered data
        forw = buff[:-step * 2, :]
        back = buff[step * 2:, :]
        data += forw + back
        
    # Normalize buffered data
    if normalize:
        denom = data.sum(axis=0)
        data = np.divide(data, denom, out=np.zeros_like(data), where=denom!=0)
    
    return data


from shapely.geometry import LineString, MultiLineString
import numpy as np
import math


class CurveDetector(object):
    """
    Class for detecting curves along polylines based on geometric features of 
    rays, arcs, and complex segments.

    To-do
    -----
    - Deal with adjacent curves/reverse curves
    - Minimum points/length for curve definition
    - Add point/length buffer
    - Full curve radius estimate
    """

    def __init__(self, line):
        # Log parameters
        self.line = line

        # Initialize fitting properties
        self._segment_mask = None

    @property
    def line(self):
        return self._line

    @line.setter
    def line(self, line):
        if not isinstance(line, (LineString, MultiLineString)):
            raise TypeError("Input line must be valid shapely linear geometry.")
        else:
            self._line = line

    @property
    def segment_mask(self):
        """
        A boolean array mask indicating which 4-point segments have been 
        detected to be part of a fitted curve.
        """
        return self._segment_mask

    @property
    def point_mask(self):
        """
        A boolean array mask indicating which points have been detected to be 
        part of a fitted curve
        """
        size = self.size
        mask = np.zeros(size, dtype=bool)
        for i in range(4):
            mask[i:size+i-3] = mask[i:size+i-3] | self.segment_mask
        return mask

    @property
    def point_map(self):
        """
        Array of values indicating which points are associated with which 
        curves. Points with a value of 0 are not associated with any unique 
        fitted curve.
        """
        mask = self.point_mask
        res = (np.append([True], np.diff(mask * 1) > 0) * mask).cumsum() * mask
        return res

    @property
    def curves(self):
        """
        A list of shapely LineStrings for fitted curves.
        """
        # Iterate through unique curve numbers
        point_map = self.point_map
        lines = []
        for num in np.unique(point_map)[1:]:
            # Get mask for curve number
            mask = point_map == num
            # Create curve linestring
            lines.append(LineString(list(zip(self.xs[mask], self.ys[mask]))))
        return lines
            
    @property
    def size(self):
        return len(self.xs)

    @property
    def xs(self):
        # Compute x values
        xs = np.array(self.line.xy[0], dtype=float)
        return xs

    @property
    def ys(self):
        # Compute x values
        ys = np.array(self.line.xy[1], dtype=float)
        return ys

    @property
    def dx(self):
        """
        X-dimension distance between adjacent points.
        Size = n - 1
        """
        dx = np.diff(self.xs)
        return dx

    @property
    def dy(self):
        """
        Y-dimension distance between adjacent points.
        Size = n - 1
        """
        dy = np.diff(self.ys)
        return dy

    @property
    def bearing(self):
        """
        Bearing of the ray defined by two adjacent points.
        Size = n - 1
        """
        bearing = np.arctan2(self.dy, self.dx)
        return bearing

    @property
    def ray_length(self):
        """
        Length of the ray defined by two adjacent points.
        Size = n - 1
        """
        ray_length = (self.dx ** 2 + self.dy ** 2) ** 0.5
        return ray_length

    @property
    def relangle(self):
        """
        Relative angle between two adjacent rays.
        Size = n - 2
        """
        relangle = np.diff(self.bearing)
        return relangle

    @property
    def direction(self):
        """
        The direction of the relative angle between two adjacent rays, 
        indicating a left-hand angle (1) and a right-hand angle (0).
        Size = n - 2
        """
        direction = (self.relangle > 0) * 1
        return direction

    @property
    def span(self):
        """
        Span length of each 3-point arc, measuring between the begin and end 
        points of each arc defined by two adjacent rays.
        Size = n - 2
        """
        dx, dy = self.dx, self.dy
        span = ((dx[:-1]+dx[1:])**2+(dy[:-1]+dy[1:])**2)**0.5
        return span
    
    @property
    def span_ratio(self):
        """
        Ratio of the smaller ray to the larger ray of each arc defined by 
        two adjacent rays.
        Size = n - 2
        """
        ray_length = self.ray_length
        span_ratio = ray_length[:-1]/ray_length[1:]
        span_ratio = np.where(span_ratio>1, 1/span_ratio, span_ratio)
        return span_ratio

    def span_index(self, span_ratio_sensitivity=0.2):
        """
        Abstract quantification of the influence of each arc's span ratio on 
        curve detection given an input sensitivity value between 0 and 1.
        Size = n - 2
        
        Parameters
        ----------
        span_ratio_sensitivity : number [0, 1], default 0.35
            A measure of the sensitivity of the detector to a given arc's span 
            ratio where 0 means no sensitivity and 1 means full sensitivity.
        """
        span_index = self.span_ratio*span_ratio_sensitivity+(1-span_ratio_sensitivity)
        return span_index

    @property
    def radius(self):
        """
        The radius of each arc defined by two adjacent rays.
        Size = n - 2
        """
        # Compute radius based on arc span and relative angle
        radius = self.span/(2*np.sin(math.pi-self.relangle))
        # Address effective tangents
        radius = np.where(self.relangle<math.pi, radius, np.inf)
        return radius

    @property
    def central_angle(self):
        """
        The central angle of each arc defined by two adjacent rays.
        Size = n - 2
        """
        central_angle = 2*np.arcsin(self.span/(2*self.radius))
        return central_angle

    @property
    def arc_length(self):
        """
        The actual outer arc length of each arc defined by two adjacent rays.
        Size = n - 2
        """
        arc_length = self.central_angle*self.radius
        return arc_length

    @property
    def radius_max(self):
        """
        The maximum radius between all pairs of adjacent arcs (i.e., 4-point 
        segments).
        Size = n - 3
        """
        radius_max = np.max([self.radius[:-1],self.radius[1:]], axis=0)
        return radius_max

    @property
    def radius_dif(self):
        """
        The mathematical difference in radius between all pairs of adjacent 
        arcs (i.e., 4-point segments).
        Size = n - 3
        """
        radius_dif = (self.radius[1:]-self.radius[:-1])
        return radius_dif

    @property
    def radius_scale(self):
        """
        The mathematical difference in radius between all pairs of adjacent 
        arcs (i.e., 4-point segments), normalized by the maximum radius between 
        each pair.
        Size = n - 3
        """
        radius_scale = self.radius_dif/self.radius_max
        return radius_scale

    def fit(self, max_radius=10000, max_radius_scale=0.65, span_ratio_sensitivity=0.35):
        """
        A test of whether or not adjacent arcs have similar radii based on 
        input detection parameters for the maximum radius scale and the 
        maximum radius which can be detected within a curve.
        Size = n - 3
        
        Parameters
        ----------
        max_radius : number, default 10000
            The maximum radius to be considered within a detected curve.
        max_radius_scale : float [0, 1], default 0.65
            The maximum span ratio-adjusted radius scale value to be considered 
            within a detected curve. If span_ratio_sensitivity == 0, no 
            adjustements to the radius scale are made.
        span_ratio_sensitivity : number [0, 1], default 0.35
            A measure of the sensitivity of the detector to a given arc's span 
            ratio where 0 means no sensitivity and 1 means full sensitivity.
        """
        # Compute span index
        # - Span index is used to adjust sensitivity to differing radii which 
        #   may result from inconsistently spaced line points (i.e., large 
        #   span index values)
        span_index = self.span_index(span_ratio_sensitivity)
        # Compute radius match boolean mask
        self._segment_mask = \
            (np.abs(self.radius_scale*span_index[:-1]*span_index[1:]) < max_radius_scale) & \
            (np.abs(self.radius_max) < max_radius)



"""
===============================================================================

Module featuring classes and functionality for spatial analysis of events data 
including parallel projection.


Classes
-------
ParallelProjector


Dependencies
------------
pandas, numpy, copy, warnings, functools


Development
-----------
Developed by:
Tariq Shihadah, tariq.shihadah@gmail.com

Created:
3/3/2022

Modified:
3/3/2022

===============================================================================
"""


################
# DEPENDENCIES #
################

import pandas as pd
import geopandas as gpd
import numpy as np
import copy, warnings
from functools import wraps
from linref.various.geospatial import join_nearby


class ParallelProjector(object):
    """
    """

    def __init__(self, target, other, samples=3, buffer=100) -> None:
        self.target = target
        self.other = other
        self.samples = samples
        self.buffer = buffer

    @property
    def samples(self):
        return self._samples

    @samples.setter
    def samples(self, samples):
        # Validate
        if not isinstance(samples, int):
            raise ValueError("Samples parameter must be an integer.")
        self._samples = samples

        # Build sampling points
        self._build_sample_points()

    @property
    def buffer(self):
        return self._buffer

    @buffer.setter
    def buffer(self, buffer):
        # Validate
        try:
            buffer = float(buffer)
            assert buffer >= 0
            self._buffer = buffer
        except:
            raise ValueError("Buffer parameter must be a non-negative "
                             "numeric value.")
        
        # Perform spatial join
        self._buffer_join()

    @property
    def sample_locs(self):
        return np.linspace(0, 1, num=self.samples)

    @property
    def sample_points(self):
        return self._sample_points

    @property
    def projectors(self):
        return self.other.geometry.values

    def _build_sample_points(self):
        """
        Build sample points along each projector geometry for matching.
        """
        # Generate sampling points
        sample_locs = self.sample_locs
        points = []
        for projector in self.projectors:
            # Interpolate sample points
            points.extend([projector.interpolate(loc) for loc in \
                sample_locs * projector.length])
        self._sample_points = gpd.GeoDataFrame({
            '__projector': np.repeat(self.other.index.values, self.samples),
            'geometry': points}, geometry='geometry', crs=self.other.crs
        )

    def _buffer_join(self):
        """
        Join projector sample points to target geometry within buffer for 
        matching.
        """
        # Join sampling points with target data
        joined = join_nearby(
            self.target.df, self.sample_points, buffer=self.buffer, choose='all'
        )
        # Process, clean results
        joined.index.rename('__target', inplace=True)
        joined = joined \
            .reset_index(drop=False)[['__projector','__target','DISTANCE']]
        joined = joined \
            .sort_values(by=['__projector','__target'], ascending=True)
        joined = joined.dropna(how='any')
        self._joined = joined

    def match(self, match='all', choose=1, sort_locs=True):
        """
        """
        # Validate matching parameters
        if match=='all':
            match = self.samples
        elif not isinstance(match, int):
            raise ValueError("Match parameter must be 'all' or an integer <= "
                             "samples.")
        # Validate choose parameter
        if not isinstance(choose, int):
            if not choose=='all':
                raise ValueError("Choose parameter must be 'all' or an "
                                "integer >= 1")
        elif choose < 1:
            raise ValueError("Integer choose parameter must be >= 1")
        # Get target event bound labels
        labels = [self.target.beg, self.target.end]

        # Group unique pairs of targets and projectors
        pair_unique, pair_index, pair_counts = np.unique(
            self._joined.values[:,:2].astype(int),
            axis=0,
            return_index=True,
            return_counts=True
        )
        # Test all unique pairs for minimum match count
        match_mask = pair_counts >= match
        # Compute mean distances for all unique matched pairs
        split = np.array(np.split(self._joined.values[:,2], pair_index)[1:])\
            [match_mask]
        mean_distances = np.array([np.mean(i) for i in split])
        # Group matched targets for all matched projectors
        proj_unique, proj_index = np.unique(
            pair_unique[match_mask,0],
            axis=0,
            return_index=True
        )
        pair_distances = np.array(np.split(mean_distances, proj_index)[1:])
        
        # Identify the index of the target(s) with the lowest mean distance for 
        # each projector if requested
        pair_groups = np.split(pair_unique[match_mask,1], proj_index)[1:]
        if choose == 'all':
            pair_select = [slice(None) for i in pair_distances]
        elif choose == 1:
            pair_select = [np.argmin(i) for i in pair_distances]
        else:
            pair_select = [np.argpartition(i, min(choose, i.size)-1) \
                        [:min(choose, i.size)] \
                        for i in pair_distances]
        # - Produce a projector-target map
        zipped = zip(pair_groups, pair_select)
        targets = np.array([group[index] for group, index in zipped])
        # - Flatten map if multiple choices
        if choose != 1:
            projectors = np.repeat(proj_unique, [len(i) for i in targets])
            targets = np.concatenate(targets, axis=0)
        else:
            projectors = proj_unique

        # Select the matched pairs
        matched_pairs = pd.DataFrame({
            '__projector': projectors, '__target': targets})
        
        # Merge matched records
        proj_lines = self.other.geometry.rename('__proj_lines')
        select = matched_pairs \
            .merge(proj_lines, left_on='__projector', right_index=True)
        target_data = self.target.df[self.target.keys + [self.target.route]]
        select = select \
            .merge(target_data, left_on='__target', right_index=True)
        select = select.reset_index(drop=True)

        # Project ends onto matched routes
        def _project(route, line):
            try:
                # Project bounds onto target route
                boundary = line.boundary
                beg, end = boundary[0], boundary[-1]
                beg_loc, end_loc = route.project(beg), route.project(end)
                return beg_loc, end_loc
            except (AttributeError, IndexError):
                return np.nan, np.nan
        proj_bounds = np.asarray(list(map(
            _project,
            select[self.target.route].values,
            select['__proj_lines'].values
        )))
            
        # Merge with input and target data
        if sort_locs:
            select[labels[0]] = proj_bounds.min(axis=1)
            select[labels[1]] = proj_bounds.max(axis=1)
        else:
            select[labels[0]] = proj_bounds[:, 0]
            select[labels[1]] = proj_bounds[:, 1]
        clean = self.other.drop(
            columns=labels + ['geometry','route'], errors='ignore')
        select = select \
            .merge(clean, how='left', left_on='__projector', right_index=True) \
            .drop(columns=['__projector','__target','__proj_lines','route'],
                  errors='ignore')
        return select


"""
===============================================================================

Module featuring classes and functionality for unifying events collections.


Classes
-------
EventsUnion


Dependencies
------------
pandas, numpy, copy, warnings, functools


Development
-----------
Developed by:
Tariq Shihadah, tariq.shihadah@gmail.com

Created:
4/13/2022

Modified:
4/13/2022

===============================================================================
"""


################
# DEPENDENCIES #
################

import pandas as pd
import numpy as np
import copy, warnings
from functools import wraps
from rangel import RangeCollection


class EventsUnion(object):
    """
    Parameters
    ----------
    objs : list-like of EventsCollection instances
        A selection of EventsCollection object instances to be combined 
        into a single instance based on the input parameters.
    **kwargs
        Keyword arguments to be passed to the initialization function for 
        the new EventsCollection instance.
    """

    def __init__(self, objs, **kwargs):
        self.objs = objs

    @property
    def objs(self):
        return self._objs

    @objs.setter
    def objs(self, objs):
        # Validate input object types
        if check_compatibility(objs):
            self._objs = objs

    @property
    def num_objs(self):
        return len(self.objs)

    @property
    def group_keys_unique(self):
        return list(set(
            key for obj in self.objs for key in obj.group_keys_unique))
    
    @property
    def num_keys(self):
        return self.objs[0].num_keys

    def get_groups(self, keys, empty=True):
        """
        Retrieve unique groups of events from each related collection based on 
        provided key values.

        Parameters
        ----------
        keys : key value, tuple of key values, or list of the same
            If only one key column is defined within the collections, a single 
            column value may be provided. Otherwise, a tuple of column values 
            must be provided in the same order as they appear in self.keys. To 
            get multiple groups, a list of key values or tuples may be 
            provided.
        empty : bool, default True
            Whether to allow for empty events groups to be returned when the 
            provided keys are valid but are not associated with any actual 
            events. If False, these cases will return a KeyError.
        """
        # Retrieve groups from all collections
        groups = [obj.get_group(keys, empty=empty) for obj in self._objs]
        return groups

    def union(
        self, 
        fill_gaps=False, 
        get_index=True, 
        merge=False, 
        suffixes=None, 
        **kwargs
        ):
        """
        Combine multiple EventsCollection instances into a single instance, 
        creating least common intervals among all collections and maintaining 
        all event attributes. The resulting combined events will be used to 
        create and return an EventsCollection modeled after the first indexed 
        collection in self.objs.

        Parameters
        ----------
        fill_gaps : bool, default False
            Whether to fill gaps in the merged collection with empty events. 
            These events would not be associated with any parent collection and 
            would not be populated with any events attributes.
        get_index : bool, default True
            Whether to produce columns relating each new record to the index of 
            the originating record in the input events dataframes. When this is 
            not necessary, setting to False may produce significant time 
            savings.
        merge : bool, default False
            Whether to merge columns from each original dataframe to the newly 
            created resegmented events collection dataframe. If not done during 
            the union, it can be done later by merging on the new 'index_i' 
            columns which correlate with the indices of the original 
            dataframes. To perform this merge manually, the get_index parameter 
            should be True.
        suffixes : list-like, default ['_0', ..., '_n']
            Sequence of length equal to the number of events collections being 
            unified, where each element is a string indicating the suffix to 
            add to overlapping column names in each corresponding events 
            dataframe. All entries must be unique.
        """
        # Validate suffixes
        if suffixes is None:
            suffixes = [f'_{i}' for i in range(self.num_objs)]
        else:
            try:
                assert len(suffixes) == self.num_objs
                assert len(set(suffixes)) == self.num_objs
                assert all(isinstance(suffix, str) for suffix in suffixes)
            except:
                raise ValueError(
                    "Input suffixes must be list-like of unique strings with "
                    "a length equal to the number of events collections being "
                    f"unified ({self.num_objs:,.0f}).")
        # Initialize new linear referencing data columns
        keys = []
        begs = []
        ends = []
        indices = []
        # Iterate over all unique group keys across all collections
        # For collections that do not contain a given group key, the resulting 
        # data will be left as null
        for group_key in self.group_keys_unique:
            # Get each group associated with the selected key across all 
            # collections being analyzed
            groups = self.get_groups(group_key, empty=True)
            # Retrieve the range data associated with each group being unified
            ranges = [group.rng for group in groups]
            # Union ranges
            if get_index:
                rc, index = RangeCollection.union(
                    ranges, fill_gaps=fill_gaps, return_index=True,
                    null_index=-1)
                # Reshape index arrays
                arrs = []
                for i, arr_i in enumerate(index):
                    try:
                        arr_i = np.where(
                            arr_i!=-1, groups[i].df.index.values[arr_i], np.nan)
                    except IndexError:
                        pass
                    arrs.append(arr_i)
                # Concatenate selected indices
                index = np.array(arrs).T
                indices.append(index)
            else:
                rc = RangeCollection.union(
                    ranges, fill_gaps=fill_gaps, return_index=False,
                    null_index=-1)
            # Log unified range results
            keys.append(np.tile(group_key, (rc.num_ranges, 1)))
            begs.append(rc.begs)
            ends.append(rc.ends)
            
        # Prepare resulting unified dataframe
        keys = np.concatenate(keys, axis=0)
        begs = np.concatenate(begs)
        ends = np.concatenate(ends)
        indices = np.concatenate(indices, axis=0)
        if get_index:
            indices[indices==-1] = np.nan
            data = pd.DataFrame({
                **{col: arr for col, arr in zip(self.objs[0].keys, keys.T)},
                **{self.objs[0].beg: begs, self.objs[0].end: ends},
                **{f'index_{i}': arr for i, arr in enumerate(indices.T)}
            })
        else:
            data = pd.DataFrame({
                **{col: arr for col, arr in zip(self.objs[0].keys, keys.T)},
                **{self.objs[0].beg: begs, self.objs[0].end: ends},
            })
        
        # Merge resegmented data with original dataframe columns
        if merge and get_index:
            for i, obj in enumerate(self.objs):
                suffixes_i = (None, suffixes[i])
                data = data.merge(
                    obj.df.drop(columns=self.objs[0].targets, errors='ignore'),
                    how='left', left_on=f'index_{i}', right_index=True, 
                    suffixes=suffixes_i, **kwargs)
            
        # Convert to events collection in the model of the first collection
        ec = self.objs[0].from_similar(data, geom=None)
        return ec


#####################
# LATE DEPENDENCIES #
#####################

from linref.events.collection import EventsCollection, check_compatibility


from linref.events.collection import EventsCollection, EventsGroup, from_standard
from linref.events.merge import EventsMerge, EventsMergeAttribute, EventsMergeTrace
from linref.events.union import EventsUnion
from linref.route import MLSRoute, combine_mpgs



import geopandas as gpd
from shapely.ops import nearest_points


def join_nearby(left, right, buffer=100, lsuffix='left', rsuffix='right', 
                choose='min', dist_label='DISTANCE'):
    """
    Join a geodataframe to the nearest, furthest, or all records from a second 
    geodataframe within a defined buffer area.

    Modified: 2020-09-21
    """
    # Collect geodataframe information
    index_left_orig  = left.index.name
    index_right_orig = right.index.name
    index_left  = 'index_left'  if not index_left_orig  else index_left_orig
    index_right = 'index_right' if not index_right_orig else index_right_orig

    # Perform spatial join with the given buffer
    joined = gpd.sjoin(left.set_geometry(left.geometry.buffer(buffer)), right,
                       how='left', lsuffix=lsuffix, rsuffix=rsuffix)

    # Restore original geometry and index names
    joined.geometry = left.geometry
    joined = joined.rename(
        columns={'index_left': index_left, 'index_right': index_right})

    # Define distance computation function
    def _get_distance(o1, o2):
        # If geometries are valid, compute distance
        try:
            p1, p2 = nearest_points(o1, o2)
            return p1.distance(p2)
        except AttributeError:
            return -1
            
    # Compute distances
    geoms = joined[[left.geometry.name,index_right]].merge(
        right.geometry, left_on=index_right, right_index=True, how='left')
    points = zip(geoms.iloc[:,0], geoms.iloc[:,2])
    joined[dist_label] = list(map(lambda x: _get_distance(*x), points))
    
    # Choose resulting records
    joined = joined.rename_axis(index=index_left) \
        .sort_values(by=[index_left,dist_label], ascending=[True,True])
    joined = joined.rename_axis(index=index_left_orig)
    if choose == 'min':
        joined = joined[~joined.index.duplicated(keep='first')]
    elif choose == 'max':
        joined = joined[~joined.index.duplicated(keep='last')]
    elif choose == 'all':
        pass
    else:
        raise ValueError("Choose parameter must be 'min', 'max', or 'all'.")
    return joined




"""
===============================================================================

Developed by Tariq Shihadah
tariq.shihadah@gmail.com

Created:
10/22/2019

Modified:
4/6/2022

===============================================================================
"""


################
# DEPENDENCIES #
################


from xml.dom.minidom import Attr
import numpy as np
from shapely.geometry import LineString, MultiLineString
import shapely
from rangel import RangeCollection
import copy, math


#############
# MLS ROUTE #
#############

class MLSRoute(object):
    """
    An object class to manage route mile-post information for each vertex of a
    shapely MultiLineString (MLS). An MLSRoute object instance will be able to
    convert between the actual linear distance along an MLS (starting from the
    origin of the MLS and moving downstream along the line, ignoring spaces 
    between individual LineStrings) to the route distance along the MLS based
    on provided route break information.
    
    Parameters
    ----------
    mls : shapely MultiLineString
        The linear geometry being represented by the route object.
    rte_breaks : list of lists of numerical values
        Numerical information representing the route distance values at each
        vertice of the route MultiLineString. To include breaks in route 
        values, input a separate list for each contiguous group of vertices. 
        Each list should have a number of elements equal to the number of 
        vertices in each LineString contained in the MultiLineString. If 
        rte_breaks is used, rte_ranges should not be input.
    rte_ranges : list of tuples of numerical values
        Numerical information representing the start and end distance values
        for each LineString in the MultiLineString. If rte_ranges is used, 
        rte_breaks will be automatically computed and should not be input.
    closed : str {'left', 'right', 'both', 'neither'}, default 'both'
        Whether intervals are closed on the left-side, 
        right-side, both or neither.
    """

    # Define class options
    _closed_ops = {'left', 'right', 'both', 'neither'}
    
    def __init__(self, mls, rte_breaks=None, rte_ranges=None, closed='both',
                 **kwargs):
        # Validate input parameters
        self.closed = closed
        self.mls = mls
        self.rte_breaks = rte_breaks if rte_ranges is None else \
            self._ranges_to_breaks(rte_ranges)
            
    def __len__(self):
        return self.rte_length
    
    def __str__(self):
        return self.wkt

    def __repr__(self):
        return self.wkt
        
    @property
    def mls(self):
        return self._mls

    @mls.setter
    def mls(self, obj):
        # Validate geometry
        if not isinstance(obj, MultiLineString):
            if isinstance(obj, LineString):
                obj = MultiLineString([obj])
            else:
                raise TypeError(
                    "Input geometry must be a shapely MultiLineString or "
                    "LineString.")
        # Log geometry
        self._mls = obj
        self._wkt = None
        # Accumulate vector lengths to generate geometry breaks
        self._mls_breaks = \
            np.concatenate([[0]] + self.element_lengths).cumsum()
        # Define geometry ranges
        rc = RangeCollection.from_breaks(
            breaks=self._mls_breaks, closed=self._closed)
        try:
            rc.ends[-1] = self.mls_length # Avoid rounding error
        except IndexError:
            pass
        self._mls_ranges = rc

    @property
    def wkt(self):
        try:
            assert self._wkt is not None
            return self._wkt
        except AssertionError:
            self._wkt = self.to_wkt()
            return self._wkt

    @property
    def mls_breaks(self):
        return self._mls_breaks
        
    @property
    def mls_ranges(self):
        return self._mls_ranges
    
    @property
    def rte_breaks(self):
        return self._rte_breaks

    @rte_breaks.setter
    def rte_breaks(self, data):
        # If no data, copy geometry breaks
        if data is None:
            self._rte_breaks = self._mls_breaks.copy()
            return
        # Check shape
        try:
            # Coerce np.ndarray
            data = [np.asarray(x) for x in data]
            # Check data length
            assert all(len(r)==len(l.coords) for \
                r, l in zip(data, self._mls.geoms))
        except:
            raise ValueError(
                "Route breaks data should be list of array-like of numeric "
                "values with lengths equal to the number of vertices in "
                "each corresponding line in the provided MultiLineString.")
        # Log data
        self._rte_breaks = data
        # Define M-value ranges
        self._rte_ranges = RangeCollection.from_breaks(
            breaks=data, closed=self._closed, sort=False)
        
    @property
    def rte_ranges(self):
        return self._rte_ranges
    
    @property
    def rte_length(self):
        return self.rte_ranges.total_length
    
    @property
    def mls_length(self):
        try:
            return self._mls_length
        except AttributeError:
            self._mls_length = self.mls.length
            return self._mls_length
    
    @property
    def num_lines(self):
        try:
            return self._num_lines
        except AttributeError:
            self._num_lines = len(self.mls.geoms)
            return self._num_lines
    
    @property
    def vertices(self):
        try:
            return self._vertices
        except AttributeError:
            self._vertices = [np.insert(x.cumsum(),0,0) \
                              for x in self.element_lengths]
            return self._vertices
    
    @property
    def element_lengths(self):
        try:
            return self._element_lengths
        except AttributeError:
            self._element_lengths = self._compute_element_lengths()
            return self._element_lengths
        
    @property
    def closed(self):
        return self._closed

    @closed.setter
    def closed(self, label):
        if not label in self._closed_ops:
            raise ValueError(
                f"Closed parameter must be one of {self._closed_ops}.")
        self._closed = label
    
    @classmethod
    def from_2d_paths(cls, paths, **kwargs):
        """
        Create MLSRoute instance from a list of paths, made up of a list of 
        three-element tuples with X, Y, and range location (i.e., M-value).
        """
        # Parse input paths into LineStrings and route breakpoint information
        lines  = []
        breaks = []
        for path in paths:
            lines.append(LineString([(x[0], x[1]) for x in path]))
            breaks.append([x[2] for x in path])
        
        # Return generated MLSRoute instance based on input paths
        return cls(MultiLineString(lines), rte_breaks=breaks, **kwargs)
    
    @classmethod
    def from_lines(cls, lines, begs, ends, **kwargs):
        """
        Create an MLSRoute instance from a list of LineStrings or a single 
        MultiLineString and lists of begin and end mile post values with 
        lengths equal to the number of LineStrings within the provided 
        geometry.
        
        Parameters
        ----------
        lines : MultiLineString, LineString, or list of either
            A collection of shapely linear geometries (LineStrings or 
            MultiLineStrings) to use as the basis for the MLSRoute.
        begs : list of numeric values or a single numeric value
            A list of begin mile post values equal in length to the provided 
            lines. This correlates to a single begin mile post value for each 
            linear geometry in the provided collection. If a single mile post 
            value is provided, begin mile post values for multiple lines will 
            be linearly interpolated.
        ends : list of numeric values or a single numeric value
            A list of end mile post values equal in length to the provided 
            lines. This correlates to a single end mile post value for each 
            line in the provided collection. If a single mile post value is 
            provided, end mile post values for multiple lines will be 
            linearly interpolated.
        **kwargs
            Keyword arguments to be input in the MLSRoute constructor.
        """
        # Single geometry provided
        # - LineString
        if isinstance(lines, LineString):
            full_lines = [MultiLineString([lines])]
        # - MultiLineString
        elif isinstance(lines, MultiLineString):
            full_lines = [lines]
        # List of geometries provided
        elif isinstance(lines, (list, tuple)):
            # - LineStrings
            if all(isinstance(i, LineString) for i in lines):
                full_lines = [MultiLineString(lines)]
            # - MultiLineStrings
            elif all(isinstance(i, MultiLineString) for i in lines):
                full_lines = lines
            else:
                raise ValueError(
                    "Input lines must be all LineString or all "
                    "MultiLineString shapely objects.")
        else:
            raise TypeError(
                "Input lines must be valid shapely linear geometries or list "
                "or tuple of the same. Provided lines are of type "
                f"{type(lines)}.")
        
        # Ensure valid breaks information provided
        # - Enforce list-data
        try:
            begs = list(begs)
        except TypeError:
            begs = [begs]
        try:
            ends = list(ends)
        except TypeError:
            ends = [ends]
            
        # Check for input type
        # - Single begin/end point provided
        if len(begs) == len(ends) == 1:
            full_lines = combine_mpgs(full_lines, cls=MultiLineString)
            distributed = \
                _distribute_dimensions(full_lines, begs[0], ends[0])
            ranges = np.stack([distributed[0], distributed[1]]).T
        # - One begin/end point per MultiLineString provided
        elif len(begs) == len(ends) == len(full_lines):
            begs_new = []
            ends_new = []
            for line, beg, end in zip(full_lines, begs, ends):
                distributed = \
                    _distribute_dimensions(line, beg, end)
                begs_new.extend(list(distributed[0]))
                ends_new.extend(list(distributed[1]))
            full_lines = combine_mpgs(full_lines, cls=MultiLineString)
            ranges = np.stack([begs_new, ends_new]).T
        # - One begin/end point per LineString provided
        elif len(begs) == len(ends) == sum(len(i.geoms) for i in full_lines):
            ranges = np.stack([begs, ends]).T
            full_lines = combine_mpgs(full_lines, cls=MultiLineString)
        # - Invalid input type
        else:
            raise ValueError(
                "Must provide a number of begin and end mile post values "
                "equal to the number of lines if providing multiple values. "
                f"Provided: {len(begs):,.0f} begs, {len(ends):,.0f} ends, "
                f"{len(mls.geoms):,.0f} lines.")
        
        # Return generated MLSRoute instance based on input parameters
        return cls(full_lines, rte_ranges=ranges, **kwargs)

    @classmethod
    def from_wkt(cls, wkt, **kwargs):
        """
        Create an MLSRoute instance from a WKT string for a MULTILINESTRING or 
        LINESTRING with three to four dimensions, with the last dimension 
        being interpreted as M values.
        
        Parameters
        ----------
        wkt : str
            WKT string representing a MULTILINESTRING or LINESTRING with three 
            to four dimensions, with the last dimension representing the 
            geometry's M values.
        **kwargs
            Keyword arguments to be input in the MLSRoute constructor.
        """
        # Simplify text and validate content
        if 'MULTILINESTRING' in wkt:
            simplified = wkt.split('((')[1] \
                .replace('),',';').replace('(','').replace(')','')
        elif 'LINESTRING' in wkt:
            simplified = wkt.split('(')[1].replace(')','')
        else:
            raise ValueError(
                "Provided WKT must represent multilinestring or linestring "
                "data.")
        # Iterate through groups of coordinates to parse M values
        data = []
        breaks = []
        for coord_group in simplified.split(';'):
            # Iterate through individual coordinates in the group
            data_group = []
            breaks_group = []
            for coord in coord_group.split(','):
                try:
                    coords = [float(x) for x in coord.strip().split(' ')]
                    dims = len(coords)
                    assert (dims > 2) & (dims < 5)
                except AssertionError:
                    raise ValueError(
                        "Input WKT must have between three and four "
                        "dimensions.")
                data_group.append(coords[:-1])
                breaks_group.append(coords[-1])
            data.append(data_group)
            breaks.append(breaks_group)

        # Create MLSRoute
        return MLSRoute(MultiLineString(data), rte_breaks=breaks, **kwargs)

    @classmethod
    def concatenate(cls, routes, **kwargs):
        """
        Combine a list of MLSRoute objects into a single MLSRoute.

        NOTE:
        - MLSRoute objects will be concatenated in the order they are 
          provided.
        - Behavior of this method under non-trivial conditions has not been 
          tested.
        """
        # Validate input
        try:
            for route in routes:
                assert isinstance(route, cls)
        except:
            raise TypeError(
                "Input routes must be list-like of MLSRoute class instances.")
        
        # Combine route breaks
        rte_breaks = \
            [rte_break for route in routes for rte_break in route.rte_breaks]
        # Combine MLS geometries
        mls = combine_mpgs(\
            [route.mls for route in routes], cls=MultiLineString)
        # Generate MLS Route instance
        mr = cls(mls, rte_breaks=rte_breaks, **kwargs)
        return mr

    def _compute_element_lengths(self):
        """
        Get individual lengths of each linear element of the route 
        MultiLineString.
        """
        # Store all vectors
        lengths = []
        for line in self.mls.geoms:
            # Log vector length
            coords = line.coords
            lengths.append(np.asarray([LineString(coords[i-1:i+1]).length \
                                       for i in range(1, len(coords))]))
        return lengths
    
    def _ranges_to_breaks(self, data):
        """
        Convert begin and end range data to breaks for the generation of an 
        MLSRoute instance.
        """
        # Confirm valid input route values
        if not data is None:
            # Coerce as numpy array, check shape
            try:
                data = np.asarray(data)
                assert data.shape == (len(self._mls.geoms), 2)
            except:
                raise ValueError(
                    "Input ranges should be array-like of numeric values with "
                    "a shape of (n,2) where n equals the number of lines in "
                    "the provided MultiLineString.")
        # Convert to breaks
        all_breaks = []
        for lengths, rng in zip(self.element_lengths, data):
            try:
                delta = rng[-1] - rng[0]
            except IndexError:
                raise IndexError(
                    "Input route ranges information must be provided as a "
                    "list of tuples of start and end values.")
            lengths = lengths / lengths.sum() # Normalize
            lengths = (lengths.cumsum() * delta) + rng[0]
            lengths[-1] = rng[-1] # Snap last value to range bound
            all_breaks.append(np.concatenate([rng[0], lengths], axis=None))
        return all_breaks

    def to_wkt(self, decimals=None):
        """
        Produce a WKT string representing the object with the underlying 
        MultiLineString appended with M values represented in the rte_breaks 
        property.
        """
        # Define number formatter
        if decimals is None:
            fmt = lambda x: str(x)
        elif isinstance(decimals, int):
            fmt = lambda x: '{x:.{decimals:.0f}f}' \
                .format(decimals=decimals, x=float(x))
        else:
            raise ValueError("Decimals parameter must be an integer.")

        # Retrieve the base WKT for the multilinestring
        wkt = self.mls.wkt
        simplified = \
            wkt.split('((')[1].replace('),',';').replace('(','').replace(')','')
        # Iterate through groups of coordinates to append M values
        data = []
        zipped = zip(simplified.split(';'), self.rte_breaks)
        for coord_group, m_group in zipped:
            # Iterate through individual coordinates in the group
            data_group = []
            for coord, m in zip(coord_group.split(','), m_group):
                point = ' '.join(fmt(x) for x in coord.strip().split(' ') + [m])
                data_group.append(point)
            data.append('(' + ', '.join(data_group) + ')')

        # Determine WKT prefix
        prefix = wkt.split(' (')[0]
        if prefix[-1] == 'Z':
            prefix += 'M '
        else:
            prefix += ' M '
        # Return combined WKT string
        return prefix + '(' + ', '.join(data) + ')'
    
    def copy(self, deep=False):
        """
        Create an exact copy of the MLS route object instance.
        """
        if deep:
            return copy.deepcopy(self)
        else:
            return copy.copy(self)
        
    def locate_mls(self, loc, normalized=False, choose='first',
                   bounded=False):
        """
        Get the range index and the proportional distance along that range
        of the input MLS location.
        
        Parameters
        ----------
        
        ...
        bounded : boolean, default False
            Whether to raise an error when the location information falls
            outside the minimum and maximum bounds of the MLS range. If False,
            negative loc values will be snapped to 0 and loc values greater 
            than mls_length will be snapped to mls_length. If True, a 
            ValueError will be raised for values which fall outside of this 
            range.
        """
        # Convert normalized values to MLS values
        if normalized:
            loc = loc * self.mls_length
            
        # Validate input
        if loc < 0:
            if bounded:
                raise ValueError(
                    "Location value cannot be negative when the bounded "
                    "parameter is True. Change to False to snap negative "
                    "values to zero.")
            else:
                loc = 0
        elif loc > self.mls_length:
            if bounded:
                raise ValueError(
                    "Location value cannot be greater than the total length "
                    "of the MLS when the bounded parameter is True. Change "
                    "to False to snap negative values to the total length of "
                    "the MLS.")
            else:
                loc = self.mls_length
        
        # Compute and return the index and proportional distance
        return self.mls_ranges.locate(loc, choose=choose)
        
    def locate_rte(self, loc, normalized=False, choose='first', snap=None):
        """
        Get the range index and the proportional distance along that range
        of the input route location.

        Parameters
        ----------
        snap : {None, 'near', 'left', 'right'}, default None
            If the input location does not fall within any ranges, snap to the 
            nearest match based on distance, choosing the closest range to the 
            left, right, or either side ('near'). If None, a value error will 
            be raised when no intersecting ranges are found.
        """
        # Convert normalized values to MLS values
        if normalized:
            loc = loc * self.mls_length
            # Compute and return the index and proportional distance
            return self.mls_ranges.locate(loc, choose=choose, snap=snap)
        else:
            # Compute and return the index and proportional distance
            return self.rte_ranges.locate(loc, choose=choose, snap=snap)
    
    def normalize(self, loc, by_mls=False, snap=None, **kwargs):
        """
        Normalize a location as an actual route location or the absolute 
        distance along the route's MultiLineString.
        
        Parameters
        ----------
        loc : numerical
            The distance along the route which will be normalized. Can be 
            route location or MultiLineString distance.
        by_mls : boolean, default False
            Whether to interpret the provided location along the route in 
            terms of the actual cumulative length of the route's 
            MultiLineString. If False, interpret as route location. If True, 
            interpret as MLS location.
        snap : {None, 'near', 'left', 'right'}, default None
            If the input location does not fall within any ranges, snap to the 
            nearest match based on distance, choosing the closest range to the 
            left, right, or either side ('near'). If None, a value error will 
            be raised when no intersecting ranges are found.
        """
        # Convert to MLS if required
        if not by_mls:
            loc = self.convert_to_mls(
                loc, normalized=False, snap=snap, **kwargs)
        # Normalize location
        res = loc / self.mls_length
        return res
    
    def convert(self, mls_loc=None, rte_loc=None, choose='first'):
        """
        Convert an mls location to a reference location or vice versa.
        """
        if not mls_loc is None:
            return self.convert_to_rte(mls_loc, choose=choose)
        elif not rte_loc is None:
            return self.convert_to_mls(rte_loc, choose=choose)
        else:
            raise ValueError("No locator inputs provided.")
        
    def convert_to_rte(self, loc=None, normalized=False, choose='first',
                       bounded=False):
        """
        Convert an MLS or normalized reference location to a route location.
        
        Parameters
        ----------
        loc : numerical
            The location along the route in terms of the absolute distance 
            along the route's MultiLineString.
        normalized : boolean, default False
            Whether to interpret the provided location along the route in 
            terms of proportional distance along the route. If False, the 
            location along the route will be interpreted normally.
        choose : {'first', 'last', 'all'}, default 'first'
            Which range to return information for if multiple ranges are found 
            which intersect with the provided location.
        bounded : boolean, default False
            Whether to raise an error when the location information falls
            outside the minimum and maximum bounds of the MLS range. If False,
            negative loc values will be snapped to 0 and loc values greater 
            than mls_length will be snapped to mls_length. If True, a 
            ValueError will be raised for values which fall outside of this 
            range.
        """
        # Locate the reference value on the route
        reference = self.locate_mls(
            loc=loc, normalized=normalized, choose=choose, bounded=bounded)
        return self.rte_ranges.project(*reference)
        
    def convert_to_mls(self, loc=None, normalized=False, choose='first',
                       snap=None):
        """
        Convert a route or normalized reference location to an MLS location.
        
        Parameters
        ----------
        loc : numerical
            The location along the route in terms of route's defined location 
            values.
        normalized : boolean, default False
            Whether to interpret the provided location along the route in 
            terms of proportional distance along the route. If False, the 
            location along the route will be interpreted normally.
        choose : {'first', 'last', 'all'}, default 'first'
            Which range to return information for if multiple ranges are found 
            which intersect with the provided location.
        snap : {None, 'near', 'left', 'right'}, default None
            If the input location does not fall within any ranges, snap to the 
            nearest match based on distance, choosing the closest range to the 
            left, right, or either side ('near'). If None, a value error will 
            be raised when no intersecting ranges are found.
        """
        # Locate the mls value on the route
        reference = self.locate_rte(loc=loc, normalized=normalized,
                                    choose=choose, snap=snap)
        return self.mls_ranges.project(*reference)
        
    def project(self, obj, by_mls=False, normalized=False):
        """
        Find the location along the route to a point nearest the input object. 
        This can be done using a normalized proportional distance along the 
        route, or the distance along the route in terms of route location or 
        MultiLineString absolute length.
        
        Parameters
        ----------
        obj : shapely geometry object
            A shapely geometry object to be projected along the route.
        by_mls : boolean, default False
            Whether to return the projected distance along the route in terms 
            of the actual cumulative length of the route's MultiLineString. If 
            False, interpret as route locations. If True, interpret as MLS 
            location. If the normalized parameter is True, this will be 
            superseded and proportional distance will be used.
        normalized : boolean, default False
            Whether to return the projected distance in terms of proportional 
            distance along the route. If False, the distance along the route 
            will be returned according to the by_mls parameter.
        """
        # Project along the MultiLineString
        loc = self.mls.project(obj, normalized=True)
        # Convert to requested projection
        if normalized:
            return loc
        else:
            if by_mls:
                return self.convert_to_mls(loc=loc, normalized=True)
            else:
                return self.convert_to_rte(loc=loc, normalized=True)
            
    def interpolate(self, loc, by_mls=False, normalized=False, snap=None, 
            **kwargs):
        """
        Return a point at the specified location along the route. This can be 
        done using a normalized proportional distance along the route, or the 
        distance along the route in terms of route location or MultiLineString 
        absolute length.

        Parameters
        ----------
        loc : numerical
            The distance along the route at which to create the point. Can be 
            proportional distance, route location, or MultiLineString 
            distance.
        by_mls : boolean, default False
            Whether to interpret the provided location along the route in 
            terms of the actual cumulative length of the route's 
            MultiLineString. If False, interpret as route locations. If True, 
            interpret as MLS location. If the normalized parameter is True, 
            this will be superseded and proportional distance will be used.
        normalized : boolean, default False
            Whether to interpret the provided location along the route in 
            terms of proportional distance along the route. If False, the 
            location along the route will be interpreted according to the 
            by_mls parameter.
        snap : {None, 'near', 'left', 'right'}, default None
            If the input location does not fall within any ranges, snap to the 
            nearest match based on distance, choosing the closest range to the 
            left, right, or either side ('near'). If None, a value error will 
            be raised when no intersecting ranges are found.
        """
        # Convert to mls reference
        if normalized:
            loc = self.convert_to_mls(loc, normalized=True, snap=snap)
        elif not by_mls:
            loc = self.convert_to_mls(loc, normalized=False, snap=snap)
        # Interpolate along the MultiLineString
        point = self.mls.interpolate(loc, normalized=False)
        return point
    
    def cut(self, beg, end, by_mls=False, normalized=False):
        """
        Cut the MLS route at the given begin and end points. This can be done 
        in terms of the route measure information (by_mls=False), in terms of
        MultiLineString actual cumulative length (by_mls=True), or in terms of 
        proportional distances along the route (normalized=True).
        
        Parameters
        ----------
        beg : float
            The location value at which the new route should begin.
        end : float
            The location value at which the new route should end.
        by_mls : boolean, default False
            Whether to interpret the begin and end points in terms of the 
            actual cumulative length of the route's MultiLineString. If False,
            interpret as route locations. If True, interpret as MLS locations. 
            If the normalized parameter is True, this will be superseded and 
            proportional distance will be used.
        normalized : boolean, default False
            Whether to interpret the begin and end points in terms of 
            proportional distances along the route. If False, the begin and 
            end points will be interpreted according to the by_mls parameter.
        
        Returns
        -------
        route : MLSRoute
            A new MLSRoute object instance with route information and a
            MultiLineString which has been cut according to the given 
            parameters.
        """
        # Validate input
        try:
            # Ensure positive numeric values
            beg = max(float(beg), 0)
            end = max(float(end), 0)
        except:
            raise ValueError("Invalid begin or end input values.")
        
        # Convert to MLS locations
        if normalized:
            beg = beg * self.mls_length
            end = end * self.mls_length
        elif not by_mls:
            beg = self.convert_to_mls(
                loc=beg, normalized=False, choose='first', snap='right')
            end = self.convert_to_mls(
                loc=end, normalized=False, choose='first', snap='left')

        # Interpolate begin point and compute range index of begin point
        if not beg is None:
            beg_point = self.interpolate(beg, by_mls=True, 
                                         normalized=False).coords[0]
            beg_index, beg_dist = self.mls_ranges.locate(loc=beg,
                             choose='last', closed='both', snap='right')
            beg_loc = self.rte_ranges.project(beg_index, beg_dist)
        else:
            beg_point = self.mls[0].coords[0]
            beg_index = 0
            beg_loc = self.rte_ranges.begs[0]

        # Interpolate end point and compute range index of end point
        if not end is None:
            end_point = self.interpolate(end, by_mls=True, 
                                         normalized=False).coords[0]
            end_index, end_dist = self.mls_ranges.locate(loc=end,
                             choose='first', closed='both', snap='left')
            end_loc = self.rte_ranges.project(end_index, end_dist)
        else:
            end_point = self.mls[-1].coords[-1]
            end_index = self.mls_ranges.num_ranges - 1
            end_loc = self.rte_ranges.ends[-1]
        
        # Unique case: begin and end points both fall between the same two
        # vertices
        if beg_index == end_index:
            mls = MultiLineString([LineString([beg_point, end_point])])
            breaks = [[beg_loc, end_loc]]
            return MLSRoute(mls, rte_breaks=breaks, closed=self.closed)
        
        # Construct new MultiLineString based on new begin and end points
        total_size = 0
        lines = []
        breaks = []
        
        # Iterate over the LineStrings in the MLS
        for num, (line, breaks_all) in enumerate(zip(self.mls.geoms,
                 self.rte_breaks)):
            
            # Get the number of ranges in the LineString
            points_all = list(line.coords)
            size = len(points_all)
            
            # Compute the start slicing parameter if the LineString is to be 
            # included
            if beg_index >= total_size + size - 1:
                # Not included, upstream of cut: skip
                total_size += size - 1
                continue
            elif beg_index >= total_size:
                # Included: cut within line
                i = beg_index - total_size + 1
            else:
                # Included: not cut within line
                i = None
                    
            # Compute the stop slicing parameter if the LineString is to be 
            # included
            if end_index < total_size:
                # Not included, downstream of cut: break
                break
            elif end_index < total_size + size - 1:
                # Included: cut within line
                j = end_index - total_size + 1
            else:
                # Included: not cut within line
                j = None
            
            # Collect the valid points based on slicing computations
            if i is None:
                points_select = []
                breaks_select = []
            else:
                points_select = [beg_point]
                breaks_select = [beg_loc]
            slicer = slice(i,j)
            points_select += points_all[slicer]
            breaks_select += breaks_all.tolist()[slicer]
            if j is None:
                lines.append(LineString(points_select))
                breaks.append(np.asarray(breaks_select))
                total_size += size - 1
                continue
            else:
                points_select += [end_point]
                breaks_select += [end_loc]
                lines.append(LineString(points_select))
                breaks.append(breaks_select)
                break
        
        # Create MLS route based on computed results
        mls = MultiLineString(lines)
        return MLSRoute(mls, rte_breaks=breaks, closed=self.closed)
    
    def segment(self, cuts, by_mls=False, normalized=False, **kwargs):
        """
        Cut the MLS Route into segments based on the given cut points.
        
        Parameters
        ----------
        cuts : array-like
            Array-like of numeric values representing the locations at which 
            to cut the route, either in terms of MultiLineString distance or 
            defined route locations.
        by_mls : boolean, default False
            Whether to interpret the provided location along the route in 
            terms of the actual cumulative length of the route's 
            MultiLineString. If False, interpret as route locations. If True, 
            interpret as MLS location. If the normalized parameter is True, 
            this will be superseded and proportional distance will be used.
        normalized : boolean, default False
            Whether to interpret the provided location along the route in 
            terms of proportional distance along the route. If False, the 
            location along the route will be interpreted according to the 
            by_mls parameter.
        
        Returns
        -------
        segments : list of MLSRoutes
            A list of new MLSRoute object instances, each with route 
            information and a MultiLineString which has been cut according to 
            the given parameters.        
        """
        # Validate inputs
        try:
            cuts = np.sort(np.asarray(cuts))
        except:
            raise ValueError(
                "Must provide segment cuts as array-like of numeric cutting "
                "point values.")
        
        # Compute the valid ranges of the routes
        beg =  self.mls_ranges.begs.min() if by_mls \
                else self.rte_ranges.begs.min()
        end = self.mls_ranges.ends.max() if by_mls \
                else self.rte_ranges.ends.max()

        # Iterate over cut points
        segments = []
        for beg_i, end_i in zip(cuts[:-1], cuts[1:]):
            
            # Check for valid location
            if end_i < beg or beg_i > end:
                continue
            
            # Perform cut
            segment_i = self.cut(beg_i, end_i, by_mls=by_mls, 
                                 normalized=normalized)
            # Append new segment to the list of created segments
            segments.append(segment_i)
        
        return segments

    def snap(self, loc, by_mls=False, normalized=False):
        """
        Snap a provided location value to the bounds of the route based on the 
        provided parameters. If the location falls within the bounds of the 
        route, the same value will be returned.

        Parameters
        ----------
        loc : scalar
            Location value to snap to the route bounds.
        by_mls : boolean, default False
            Whether to interpret the provided location along the route in 
            terms of the actual cumulative length of the route's 
            MultiLineString. If False, interpret as route location. If True, 
            interpret as MLS location.
        normalized : boolean, default False
            Whether to interpret the provided location along the route in 
            terms of proportional distance along the route. If False, the 
            location along the route will be interpreted according to the 
            by_mls parameter.
        """
        # Snap location by selected range collection
        if normalized:
            return max(min(loc, 1), 0)
        elif by_mls:
            return self.mls_ranges.snap(loc)
        else:
            return self.rte_ranges.snap(loc)
    
    def bearing(self, positive=True, invert=False):
        """
        Approximate the bearing angle of the route, based on the first and 
        last points in the route's MLS.
        
        Parameters
        ----------
        positive : bool, default True
            Whether to enforce a positive range on the computed bearing angle. 
            If True, the bearing angle will fall on the range [0,360). If 
            False, the bearing angle will fall on the range (-180,180].
        invert : bool, default False
            Whether to invert the computed bearing angle, effectively 
            reversing the direction of the route.
        """
        # Capture x and y distance between points
        x_diff = self.mls[-1].xy[0][-1] - self.mls[0] .xy[0][0]
        y_diff = self.mls[-1].xy[1][-1] - self.mls[0] .xy[1][0]
        
        # Compute bearing angle
        bearing = math.degrees(math.atan2(y_diff, x_diff))
        
        # Invert if requested
        if invert:
            bearing += 180
        
        # Enforce range
        if positive and bearing < 0:
            bearing += 360
        elif not positive and bearing > 180:
            bearing -= 360
            
        return bearing


#####################
# SUPPORT FUNCTIONS #
#####################

def combine_mpgs(objs, cls=None):
    """
    Combine multiple multipart geometries into a single multipart geometry of 
    geometry collection.
    """
    # Generate new list of individual geometries
    new = []
    for obj in objs:
        if isinstance(obj, shapely.geometry.base.BaseMultipartGeometry):
            new.extend(list(obj.geoms))
        elif isinstance(obj, shapely.geometry.base.BaseGeometry):
            new.extend([obj])
        else:
            raise TypeError("Invalid geometry type")
    # Convert list to geometry collection or provided class
    if cls is None:
        new = shapely.geometry.collection.GeometryCollection(new)
    else:
        new = cls(new)
    return new

def _distribute_dimensions(mls, beg, end):
    # Validate input
    if not isinstance(mls, MultiLineString):
        raise ValueError("Input MLS must be MultiLineString type.")
    if mls.is_empty:
        raise ValueError("Input MLS is empty.")
    # Compute dimensions
    delta = end - beg
    lengths = np.array([ls.length for ls in mls.geoms])
    proportions = np.cumsum(lengths / lengths.sum() * delta)
    begs = np.insert(proportions[:-1] + beg, 0, beg)
    ends = proportions + beg
    # Return proportions
    return begs, ends
    

# Sample use
if __name__ == '__main__':
    # Create a generic MLS
    mls = MultiLineString([[(0,0), (0,5), (5,5), (5,0)],
                           [(5,0), (5,-5), (10,-5), (10,0)],
                           [(10,0), (10,5), (15,5), (15,0)],
                           [(20,0), (20,-5), (25,-5), (25,0)]])                
                
    # Create a generic range set
    rte_ranges = [(0,150), (200,350), (400,550), (600,650)]
    
    # Create a MLS route
    route = MLSRoute(mls, rte_ranges=rte_ranges)
    
    # Cut the sample route
    test = route.cut(0,0.80, normalized=True)
        
        


"""
===============================================================================

Module featuring classes and functionality for merging events collections and 
summarizing/retrieving information from these merges. For ease of use, features 
in this module should be accessed through collection-level merging methods such 
as EventsCollection.merge instead of abstractly through the classes themselves.


Classes
-------
EventsMerge, EventsMergeAttribute, EventsMergeTrace


Dependencies
------------
pandas, numpy, rangel, copy, warnings, functools


Development
-----------
Developed by:
Tariq Shihadah, tariq.shihadah@gmail.com

Created:
10/1/2021

Modified:
10/1/2021

===============================================================================
"""


################
# DEPENDENCIES #
################

import pandas as pd
import numpy as np
import copy, warnings
from functools import wraps
from rangel import RangeCollection


class EventsMergeAttribute(object):
    
    def __init__(self, parent, column):
        self.parent = parent
        self.column = column
        
    @property
    def parent(self):
        return self._parent
    
    @parent.setter
    def parent(self, obj):
        if not isinstance(obj, EventsMerge):
            raise TypeError("Input parent must be EventsMerge type.")
        self._parent = obj
        
    @property
    def column(self):
        return self._column
    
    @column.setter
    def column(self, label):
        # Multiple labels provided
        if isinstance(label, (list, tuple, np.ndarray)):
            if not set(label).issubset(self.parent.columns):
                raise ValueError("All column labels must be present in "
                                 "EventsMerge.columns.")
            # Process input
            self._ndim = 2
            self._ncols = len(label)
            self._loc = [self.parent.right.columns.index(i) for i in label]
            self._column = list(label)
        # Single label provided
        else:
            if not label in self.parent.columns:
                raise ValueError("Invalid column label for merged events. Must "
                                 "be present in EventsMerge.columns.")
            # Process input
            self._ndim = 1
            self._ncols = 1
            self._loc = [self.parent.right.columns.index(label)]
            self._column = label
        
    @property
    def traces(self):
        try:
            return self.parent.traces
        except AttributeError:
            self.parent.build(inplace=True)
            return self.parent.traces
    
    @property
    def loc(self):
        return self._loc

    @property
    def ncols(self):
        return self._ncols

    @property
    def ndim(self):
        return self._ndim

    def _to_pandas(self, index, data, as_series=False, squeeze=True):
        # Create pandas object
        if as_series:
            obj = pd.Series(
                data=data,
                index=index,
                name=self.column
            )
        else:
            obj = pd.DataFrame(
                data=data,
                index=index,
                columns=self.column if self._ndim != 1 else [self.column]
            )
            if self._ndim == 1 and squeeze:
                obj = obj.iloc[:, 0]
        return obj

    def _get_empty(self, empty=None):
        return np.nan if empty is None else empty

    def _agg(self, func, empty=None, as_array=True, **kwargs):
        """
        Generic attribute aggregator.

        Aggregation functions will be passed a single EventsMergeTrace instance 
        for each unique EventsGroup in the left EventsCollection, along with an 
        array of the values in the right EventsGroup indexed by the selected 
        columns.
        """
        # Validate fill value
        empty = self._get_empty(empty)
        # Iterate over events to create aggregated result
        index = []
        res = []
        for trace in self.traces:
            # Where actual traces are created, perform aggregation function
            if trace.success:
                arr = trace.group_right.df.values[:, self.loc]
                res_i = func(arr, trace, **kwargs)
            # Otherwise create a full empty array
            else:
                res_i = np.full(
                    (trace.group_left.df.shape[0], self._ncols), empty)
            # Log results from trace
            index.extend(trace.group_left.df.index)
            res.append(res_i)
        # Combine results for each left group into a single array that is  
        # compatible with the left collection
        if as_array:
            return np.array(index), np.concatenate(res)
        else:
            # Flatten data (relevant for geometry aggregators)
            return index, [i for sub in res for i in sub]

    def all(self, empty=None, **kwargs):
        """
        Return all values from intersecting events in a list.

        Parameters
        ----------
        empty : scalar, string, or other pd.Series-compatible value, optional
            Value to use to fill when there is no matching events group and 
            aggregation cannot be performed. If None, values will be filled 
            with np.nan.
        """
        def _func(arr, trace, **kwargs):
            # Choose all intersecting events
            res = np.empty((trace.mask.shape[0], self._ncols), dtype=object)
            res[:] = [list(arr[mask_i, :].T) for mask_i in trace.mask]
            return res
        return self._to_pandas(*self._agg(_func, empty=empty))

    def unique(self, empty=None, **kwargs):
        """
        Return all unique values from intersecting events in a tuple.

        Parameters
        ----------
        empty : scalar, string, or other pd.Series-compatible value, optional
            Value to use to fill when there is no matching events group and 
            aggregation cannot be performed. If None, values will be filled 
            with np.nan.
        """
        def _func(arr, trace, **kwargs):
            # Choose all intersecting events
            res = np.empty((trace.mask.shape[0], self._ncols), dtype=object)
            res[:] = [[list(set(sub)) for sub in arr[mask_i, :].T] \
                for mask_i in trace.mask]
            return res
        return self._to_pandas(*self._agg(_func, empty=empty))

    def first(self, empty=None):
        """
        Return the first event value according to the order of the provided 
        collection's events dataframe.

        Parameters
        ----------
        empty : scalar, string, or other pd.Series-compatible value, optional
            Value to use to fill when there is no matching events group and 
            aggregation cannot be performed. If None, values will be filled 
            with np.nan.
        """
        def _func(arr, trace, **kwargs):
            # Identify the first potential match
            index = np.argmax(trace.mask, axis=1)
            # Identify where matches are made
            valid = trace.mask.any(axis=1).reshape(-1,1)
            # Fill empty value where no match is made
            res = np.where(valid, arr[index], self._get_empty(empty))
            return res
        return self._to_pandas(*self._agg(_func, empty=empty))
    
    def last(self, empty=None):
        """
        Return the last event value according to the order of the provided 
        collection's events dataframe.

        Parameters
        ----------
        empty : scalar, string, or other pd.Series-compatible value, optional
            Value to use to fill when there is no matching events group and 
            aggregation cannot be performed. If None, values will be filled 
            with np.nan.
        """
        def _func(arr, trace, **kwargs):
            # Identify the first potential match
            index = np.argmax(np.flip(trace.mask, axis=1), axis=1)
            # Identify where matches are made
            valid = trace.mask.any(axis=1).reshape(-1,1)
            # Fill empty value where no match is made
            res = np.where(valid, arr[index], self._get_empty(empty))
            return res
        return self._to_pandas(*self._agg(_func, empty=empty))

    def count(self, empty=None):
        """
        Return the count of all intersected event values.

        Parameters
        ----------
        empty : scalar, string, or other pd.Series-compatible value, optional
            Value to use to fill when there is no matching events group and 
            aggregation cannot be performed. If None, values will be filled 
            with np.nan.
        """
        def _func(arr, trace, **kwargs):
            # Sum count of all intersecting events, apply to all columns
            res = np.tile(trace.mask.sum(axis=1), (self._ncols,1)).T
            return res
        return self._to_pandas(*self._agg(_func, empty=empty))

    def sum(self, empty=None, nansum=False):
        """
        Return the sum of all intersected event values.

        Parameters
        ----------
        empty : scalar, string, or other pd.Series-compatible value, optional
            Value to use to fill when there is no matching events group and 
            aggregation cannot be performed. If None, values will be filled 
            with np.nan.
        """
        def _func(arr, trace, **kwargs):
            # Choose all intersecting events
            if nansum:
                res = [np.nansum(arr[mask_i, :], axis=0).T for \
                    mask_i in trace.mask]
            else:
                res = [np.sum(arr[mask_i, :], axis=0).T for \
                    mask_i in trace.mask]
            return res
        return self._to_pandas(*self._agg(_func, empty=empty))

    def sumproduct(self, empty=None, normalized=False, dropna=False):
        """
        Return the sum of all event values multiplied by the weights of the 
        intersecting events. If normalized=False, the event values will be 
        multiplied by the actual overlapping length (e.g., multiplying a per-
        mile value by the miles of overlap). If normalized=True, the event 
        values will be multiplied by the normalized overlapping length (e.g., 
        multiplying a total value of an overlapped event by the proportion of 
        the event which is overlapped).

        Parameters
        ----------
        empty : scalar, string, or other pd.Series-compatible value, optional
            Value to use to fill when there is no matching events group and 
            aggregation cannot be performed. If None, values will be filled 
            with np.nan.
        normalized : boolean, default False
            Whether the weights of the intersecting events being multiplied 
            with the event values should be normalized by the total length of 
            the events being intersected.
        dropna : boolean, default False
            Whether to drop np.nan values before aggregating.
        """
        def _func(arr, trace, **kwargs):
            # Prepare weights data congruent with array data
            weights = trace.weights
            weights = np.tile(np.expand_dims(weights, 2), (1,1,self._ncols))
            if normalized:
                weights /= trace.group_right.lengths
            # Drop nan if requested, zeroing weights where nan values occur
            if dropna:
                weights = np.where(np.isnan(arr.astype(float)), 0, weights)
           # Compute sums
            res = np.multiply(
                arr.reshape(1,-1,self._ncols), weights).sum(axis=1)
            return res
        return self._to_pandas(*self._agg(_func, empty=empty))

    def mean(self, empty=None, weighted=True, dropna=False):
        """
        Return an overlay length-weighted average of all event values. An 
        unweighted simple average can also be computed if weighted=True.

        Parameters
        ----------
        empty : scalar, string, or other pd.Series-compatible value, optional
            Value to use to fill when there is no matching events group and 
            aggregation cannot be performed. If None, values will be filled 
            with np.nan.
        weighted : boolean, default True
            Whether the computed average should be weighted. If False, an
            un-weighted average will be computed, giving all intersecting 
            values an equal weight.
        dropna : boolean, default False
            Whether to drop np.nan values before aggregating.
        """
        def _func(arr, trace, **kwargs):
            # Determine event weights
            weights = trace.weights if weighted else trace.mask
            weights = np.tile(np.expand_dims(weights, 2), (1,1,self._ncols))
            # Drop nan if requested, zeroing weights where nan values occur
            if dropna:
                weights = np.where(np.isnan(arr.astype(float)), 0, weights)
            # Compute means
            numer = np.multiply(
                arr.reshape(1,-1,self._ncols), weights).sum(axis=1)
            denom = weights.sum(axis=1)
            denom = np.where(denom==0, np.nan, denom)
            res = np.divide(numer, denom)
            return res
        return self._to_pandas(*self._agg(_func, empty=empty))

    def most(self, empty=None, dropna=True):
        """
        Return the event value associated with the greatest total overlay 
        length, ignoring missing values by default.

        Parameters
        ----------
        empty : scalar, string, or other pd.Series-compatible value, optional
            Value to use to fill when there is no matching events group and 
            aggregation cannot be performed. If None, values will be filled 
            with np.nan.
        dropna : boolean, default False
            Whether to drop np.nan values in intersecting events before 
            aggregating.
        """
        def _func(arr, trace, **kwargs):
            # Iterate over 2nd dimension
            res = []
            weights = trace.weights
            for arr_i in arr.T:
                # Drop nan if requested
                if dropna:
                    nanmask = ~pd.isna(arr_i)
                    if nanmask.sum() == 0:
                        res.append(
                            np.full(weights.shape[0], self._get_empty(empty)))
                        continue
                    else:
                        arr_i = arr_i[nanmask]
                        weights_i = weights[:,nanmask]
                else:
                    weights_i = weights
                # Aggregate and add to result
                res.append(get_most(arr_i, weights_i))
            return np.array(res).T
        return self._to_pandas(*self._agg(_func, empty=empty))
        
    def mode(self, empty=None):
        """
        Return the most frequent unique event value.

        Parameters
        ----------
        empty : scalar, string, or other pd.Series-compatible value, optional
            Value to use to fill when there is no matching events group and 
            aggregation cannot be performed. If None, values will be filled 
            with np.nan.
        """
        def _func(arr, trace, **kwargs):
            # Iterate over all event masks
            res = []
            for mask_i in trace.mask:
                # Iterate over columns
                res_i = [get_mode(arr_i) for arr_i in arr[mask_i, :].T]
                res.append(res_i)
            return np.array(res)
        return self._to_pandas(*self._agg(_func, empty=empty))

    def value_counts(self, expand=True, dropna=True):
        """
        Return a dataframe of all unique intersecting event values and their 
        occurence counts.

        Parameters
        ----------
        expand : bool, default True
            Whether to automatically expand the value counts data to a 
            dataframe when a single column is being analyzed.
        """
        def _func(arr, trace, **kwargs):
            # Find null values in array
            if dropna:
                nanmask = ~pd.isna(arr.T)
            # Iterate over all event masks
            res = []
            for mask_i in trace.mask:
                # Iterate over columns
                res_i = []
                for j in range(arr.shape[1]):
                    # Remove NaN
                    if dropna:
                        arr_i = arr[mask_i & nanmask[j]]
                    else:
                        arr_i = arr[mask_i]
                    # Get counts of unique values
                    unique, counts = np.unique(
                        arr_i, return_counts=True, equal_nan=True)
                    # Summarize in dictionary
                    res_i.append(
                        {val: count for val, count in zip(unique, counts)})
                res.append(res_i)
            return np.array(res)
        # Generate data and expand if requested
        index, data = self._agg(_func, empty={})
        if self._ndim == 1:
            if expand:
                return pd.DataFrame(data.flatten().tolist(), index=index)
            else:
                return self._to_pandas(index, data)
        else:
            return self._to_pandas(index, data)
    
    def cut(self, empty=None, return_mls=True):
        """
        Cut intersecting event routes at the intersecting begin and end 
        locations, returning the resulting route's geometry or the route itself 
        if requested.

        Parameters
        ----------
        empty : scalar, string, or other pd.Series-compatible value, optional
            Value to use to fill when there is no matching events group and 
            aggregation cannot be performed. If None, values will be filled 
            with np.nan.
        return_mls : bool, default True
            Whether to return the MultiLineString associated with each cut 
            MLRRoute instead of the route itself.
        """
        if self._ncols != 1:
            raise ValueError("EventsMergeAttribute must represent a single "
                             "column to perform cut.")
        def _func(arr, trace, **kwargs):
            # Iterate over all event masks
            indices = np.argmax(trace.mask, axis=1)
            gen = zip(indices, trace.group_left.begs, trace.group_left.ends)
            res = []
            for i, beg, end in gen:
                # Choose the first intersecting event and cut the route
                route = arr[i, 0]
                try:
                    res_i = route.cut(beg, end)
                    res_i = res_i.mls if return_mls else res_i
                    res.append(res_i)
                except AttributeError:
                    raise TypeError(
                        "EventsMergeAttribute must represent a single column "
                        "and must contain MLSRoute objects to be cut.")
            return res
        return self._to_pandas(*self._agg(
            _func, empty=empty, as_array=False), as_series=True)
    
    def interpolate(self, snap=None, point='begs', empty=None, **kwargs):
        """
        Interpolate along intersecting event routes at the intersecting 
        location (or begin point for linear events), returning the resulting 
        interpolated point geometry.

        Parameters
        ----------
        snap : {None, 'near', 'left', 'right'}, default None
            If the event location does not fall within any geometry, snap to 
            the nearest match based on distance, choosing the closest location 
            to the left, right, or the nearest side ('near'). If None, a value 
            error will be raised when no intersecting ranges are found.
        point : {'begs', 'ends', 'centers'}, default 'begs'
            Where on the intersecting events the point should be made, at the 
            begin, end, or center point of the range.
        empty : scalar, string, or other pd.Series-compatible value, optional
            Value to use to fill when there is no matching events group and 
            aggregation cannot be performed. If None, values will be filled 
            with np.nan.
        """
        if self._ncols != 1:
            raise ValueError("EventsMergeAttribute must represent a single "
                             "column to perform interpolate.")
        def _func(arr, trace, **kwargs):
            # Iterate over all event masks
            indices = np.argmax(trace.mask, axis=1)
            gen = zip(indices, getattr(trace.group_left, point))
            res = []
            for i, loc in gen:
                # Choose the first intersecting event and interpolate the point
                route = arr[i, 0]
                try:
                    res_i = route.interpolate(loc, snap=snap, **kwargs)
                    res.append(res_i)
                except AttributeError:
                    raise TypeError(
                        "EventsMergeAttribute must represent a single "
                        "column and must contain MLSRoute objects to be "
                        "interpolated.")
            return res
        return self._to_pandas(*self._agg(
            _func, empty=empty, as_array=False), as_series=True)
     

class EventsMergeTrace(object):
    """
    Object class for managing data on the relationship between two events 
    collections that have been merged using the EventsMerge system. Traces 
    contain a few main elements:

    group_left, group_right : pointers to the left and right events groups that 
        are related. During aggregation, information in the right group will be 
        aggregated and formed to the dataframe underlying the left group.
    key : the unique key associated with both events groups that produces their 
        relationship.
    mask : a boolean array of shape (group_left.df.shape[0], 
        group_right.df.shape[0]), i.e., a number of rows equal to the number of 
        rows in the left events group and a number of columns equal to the 
        number of rows in the right events group. This mask defines all 
        instances where the left and right groups intersect based on their 
        defined ranges and closed parameters.
    weights : a numeric array of shape (group_left.df.shape[0], 
        group_right.df.shape[0]), i.e., a number of rows equal to the number of 
        rows in the left events group and a number of columns equal to the 
        number of rows in the right events group. This array defines the actual 
        numeric length that is overlapped between the individual events in the 
        left and right events groups.
    success : a boolean indicator of whether or not a valid relationship has 
        been discovered between the left group and any right group. When False, 
        no right group will be indicated.
    """
    
    def __init__(self, group_left=None, group_right=None, key=None, mask=None, 
            weights=None, success=True):
        self.group_left = group_left
        self.group_right = group_right
        self.key = key
        self.mask = mask
        self.weights = weights
        self.success = success


class EventsMerge(object):
    """
    High-level object class for managing merges between two events collections 
    and summarizing/retrieving information from these merges. Generated through 
    collection-level merging methods such as EventsCollection.merge.
    """
    
    def __init__(self, left, right):
        # Log parameters
        self.left = left
        self.right = right
        
    def __getitem__(self, column) -> EventsMergeAttribute:
        return EventsMergeAttribute(self, column)
    
    def __repr__(self):
        text = (
            f"left:  {self.left}\n"
            f"right: {self.right}\n"
            f"(traces {'built' if hasattr(self, '_traces') else 'not built'})"
        )
        return text
        
    @property
    def left(self):
        return self._left
    
    @left.setter
    def left(self, obj):
        self._validate_target(obj)
        self._left = obj
        
    @property
    def right(self):
        return self._right
    
    @right.setter
    def right(self, obj):
        # Validate and set
        self._validate_target(obj, left=False)
        self._right = obj
        
    @property
    def traces(self):
        try:
            return self._traces
        except AttributeError:
            self.build(inplace=True)
            return self._traces
    
    @traces.setter
    def traces(self, obj):
        if not isinstance(obj, list):
            raise TypeError("Traces must be a list of EventsMerge_trace "
                            "objects.")
        self._traces = obj
        
    @property
    def num_keys(self):
        return self.left.num_keys
        
    @property
    def keys(self):
        return self.left.keys
    
    @property
    def columns(self):
        return self.right.columns
        
    def _validate_target(self, obj, left=True):
        # Ensure left is set first if the target is the right
        if not (left) and not (hasattr(self, '_left')):
            raise AttributeError("The left target must be set before the right "
                                 "target.")
        # Ensure type
        if not isinstance(obj, EventsCollection):
            raise TypeError("EventsMerge targets must be EventsCollections.")
        # Ensure matching keys
        try:
            assert obj.num_keys == self.num_keys
        except AttributeError:
            pass
        except AssertionError:
            raise ValueError(
                "Input EventsMerge target must have the same number of keys "
                f"as the existing left target ({self.num_keys}).")
            
    def copy(self, deep=False):
        """
        Create an exact copy of the events class instance.
        
        Parameters
        ----------
        deep : bool, default False
            Whether the created copy should be a deep copy.
        """
        if deep:
            return copy.deepcopy(self)
        else:
            return copy.copy(self)

    def build(self, inplace=True):
        """
        Perform intersects and overlays to produce EventsMergeTrace objects 
        for aggregation.
        """
        # Iterate over EventsGroups in the left collection
        traces = []
        for key, group_left in self.left.iter_groups():
            # Get the associated EventsGroup from the right collection
            try:
                group_right = self.right.get_group(key, empty=False)
            except KeyError:
                # No matching group on the right? Log empty trace
                traces.append(
                    EventsMergeTrace(
                        group_left=group_left, key=key, success=False))
                continue
            # Perform intersection
            mask = group_left.intersecting(
                other=group_right, squeeze=False, get_mask=True)
            weights = group_left.overlay(
                other=group_right, normalize=False, squeeze=False)
            # Aggregate results
            traces.append(
                EventsMergeTrace(
                    group_left=group_left, group_right=group_right, 
                    key=key, mask=mask, weights=weights, success=True))
        # Log traces
        self.traces = traces
    
    def distribute(self, column=None, squeeze=True, **kwargs):
        """
        Intersect and distribute events over the range collection, scaling 
        their values relative to their indexed distance from their intersecting 
        range location.
        
        Parameters
        ----------
        column : pandas column label or list of same, optional
            The events dataframe column(s) containing the values associated 
            with each event being analyzed. If not provided, all values will 
            default to be 1.
        blur_size : int, default 0
            The number of pixels to blur events across based on the blur style.
        blur_style : str or callable, default 'linear'
            The scaling function to be called at each blurring step to scale 
            original values. If a callable is provided, it must accept a single 
            integer input for the zero-indexed pixel number, returning a single 
            float scaling value. Predefined blurring functions can be called 
            using the following labels:
            
            Options
            -------
            linear : linearly scale down values from the original value to zero 
                at the first index outside the blurred pixel range
            norm_static : Tk
            norm_scale : Tk
            none : do not scale down original values

        length_normalize : bool, default True
            Normalize the intersection scores by the length of the range to 
            account for differing range lengths.
                
        Created:  2022-10-04
        """
        # Iterate over EventsGroups in the left collection
        index = []
        res = []
        for key, group_left in self.left.iter_groups():
            # Get the associated EventsGroup from the right collection
            group_right = self.right.get_group(key, empty=True)
            # Get event values
            if not column is None:
                if isinstance(column, list):
                    values = group_right.df[column].values
                else:
                    values = group_right.df[[column]].values
            else:
                values = None
            # Perform distribution
            res_i = group_left.rng.distribute(
                group_right.rng, values=None, **kwargs)
            # Aggregate results over all column values
            if not column is None:
                res_i = (np.expand_dims(res_i, 2) * values).sum(axis=1)
            else:
                res_i = res_i.sum(axis=1)
            # Log results
            index.append(group_left.df.index)
            res.append(res_i)
        # Synthesize into pandas object
        index = np.concatenate(index)
        data = np.concatenate(res)
        if column is None:
            obj = pd.Series(index=index, data=data, name=None)
        elif isinstance(column, list):
            if len(column)==1 and squeeze:
                obj = pd.Series(index=index, data=data, name=column[0])
            else:
                obj = pd.DataFrame(index=index, data=data, columns=column)
        else:
            obj = pd.Series(index=index, data=data, name=column)
        return obj

    def cut(self, **kwargs):
        """
        Cut intersecting event routes at the intersecting begin and end 
        locations, returning the resulting route's geometry or the route itself 
        if requested.

        Parameters
        ----------
        empty : scalar, string, or other pd.Series-compatible value, optional
            Value to use to fill when there are no intersecting events and 
            aggregation cannot be performed. If None, values will be filled 
            with np.nan.
        return_mls : bool, default True
            Whether to return the MultiLineString associated with each cut 
            MLRRoute instead of the route itself.
        """
        # Get attribute for routes if available
        try:
            ema = self[self.right.route]
        except:
            raise ValueError("Right collection does not contain a valid "
                "routes column label.")
        # Perform cut
        return ema.cut(**kwargs)

    def interpolate(self, **kwargs):
        """
        Interpolate along intersecting event routes at the intersecting 
        location (or begin point for linear events), returning the resulting 
        interpolated point geometry.

        Parameters
        ----------
        snap : {None, 'near', 'left', 'right'}, default None
            If the event location does not fall within any geometry, snap to 
            the nearest match based on distance, choosing the closest location 
            to the left, right, or the nearest side ('near'). If None, a value 
            error will be raised when no intersecting ranges are found.
        point : {'begs', 'ends', 'centers'}, default 'begs'
            Where on the intersecting events the point should be made, at the 
            begin, end, or center point of the range.
        empty : scalar, string, or other pd.Series-compatible value, optional
            Value to use to fill when there are no intersecting events and 
            aggregation cannot be performed. If None, values will be filled 
            with np.nan.
        """
        # Get attribute for routes if available
        try:
            ema = self[self.right.route]
        except:
            raise ValueError("Right collection does not contain a valid "
                "routes column label.")
        # Perform cut
        return ema.interpolate(**kwargs)

    def count(self, **kwargs):
        """
        Count the number of intersecting events.

        Parameters
        ----------
        empty : scalar, string, or other pd.Series-compatible value, optional
            Value to use to fill when there is no matching events group and 
            aggregation cannot be performed. If None, values will be filled 
            with np.nan.
        """
        # Get the first key attribute
        ema = self[self.right.keys[0]]
        # Perform aggregation
        return ema.count(**kwargs)

    def any(self, **kwargs):
        """
        Indicate whether each record intersects with at least one event.

        Parameters
        ----------
        empty : scalar, string, or other pd.Series-compatible value, optional
            Value to use to fill when there is no matching events group and 
            aggregation cannot be performed. If None, values will be filled 
            with np.nan.
        """
        # Get the first key attribute
        ema = self[self.right.keys[0]]
        # Perform aggregation
        return ema.any(**kwargs)


###########
# HELPERS #
###########

def get_most(arr, weights):
    """
    Select the item from the input 1D array which is associated with the 
    highest total weight from each row in the 2D weights array. Scores are 
    computed by summing the weights for each unique array value for each row of 
    weights. When multiple values are tied, the first item in sorted order will 
    be selected.
    """
    # Group and split sorted target array
    sorter = np.argsort(arr)
    unique, splitter = np.unique(arr[sorter], return_index=True)
    splitter = splitter[1:]
    # Split weights and aggregate to select highest scoring values
    splits = np.split(weights[:, sorter], splitter, axis=1)
    select = unique[np.argmax([x.sum(axis=1) for x in splits], axis=0)]
    return select
    
def get_mode(arr):
    """
    Select the item from the input array which appears most frequently.
    
    Parameters
    ----------
    arr : array-like
        Array with target values
    """
    # Enforce numpy array
    arr = np.asarray(arr)
    # Find most frequent unique value and return
    unique, counts = np.unique(arr, return_counts=True)
    return unique[np.argmax(counts)]


#####################
# LATE DEPENDENCIES #
#####################

from linref.events.collection import EventsCollection
